/*
This file is part of "Uakron Robot Project".

"Uakron Robot Project" is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

"Uakron Robot Project" is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "LcdLibrary.h"

//When the address of the LCD is set,
//bit7 must be a 1. This is the
//set DDRAM command. CGRAM deals with
//the custom characters
#define setLcdAddrCommand 0x80
#define line1Addr 0x0+setLcdAddrCommand
#define line2Addr 0x40+setLcdAddrCommand
#define line3Addr 0x14+setLcdAddrCommand
#define line4Addr 0x54+setLcdAddrCommand

#define clearLcdCommand 0x01



//bits 0,2 are used for the LCD.
//do PINC & clearLcdBits to clear
//these bits
#define clearLcdBits 0xFA
#define enable 0x04
//is rs bit set to 1
//rs is pin0
#define writeCommandPinCombo 0x01
#define systemCommandPinCombo 0x00
#define setAddrCommandPinCombo 0x00
//On this processor, if we run at 1MHz, that is one
//instruction per u-second. Many of the timing 
//requirements of the LCD screen are measured
//in nano-seconds. There is no need for
//sleep commands to execute commands,
//but to make sure the command is executed before
//returning, a sleep statement may be necessary.

//setup variables
#define initSetupBitPatten  0x30

#define setFunctionCommand  0x20
#define bitInterface4       0x00
#define bitInterface8       0x10
#define lineDisplay1        0x00
#define lineDisplay2        0x08
#define Font5x10            0x04
#define Font5x7             0x00

#define setEntryModeCommand 0x04
#define cursorIncreases     0x02
#define cursorDecreases     0x00
#define displayShifted      0x01
#define displayNotShifted   0x00

#define setDisplayCommand   0x08
#define turnDisplayOff      0x00
#define turnDisplayOn       0x04
#define turnCursorOn        0x02
#define turnCursorOff       0x00
#define BlinkCursor         0x01
#define noBlinkCursor       0x00





/*
 * Initializes the Shift register, then the
 * LCD display so that it can be used.
 */
void initLcd()
{
    //keep what bits are already set on DDRC, 
    //and add our bits on.
    //pin assignments (7-0):
    //bit0 is data pin for shift register,
    //   and RS pin for LCD
    //bit1 is clk for shift register
    //bit2 is Enable for LCD
    //Note: the r/w pin should be set to ground.

    DDRC = DDRC + 0x07;

    //The LCD must have at least 15ms
    //after startup to sleep. We do not know
    //how soon this operation will be executed,
    //so we will err on the side of caution.
    sleep(20);

    //must set the following bit pattern 3 times
    byte index;
    for(index = 0; index < 3; index++)
    {
        sendData(initSetupBitPatten,systemCommandPinCombo);
        sleep(10);
    }

    //set the functionality of the display
    sendData(setFunctionCommand + bitInterface8 + lineDisplay1 + Font5x10, systemCommandPinCombo);

    //turn off the display
    sendData(setDisplayCommand + turnDisplayOff + turnCursorOff + noBlinkCursor, systemCommandPinCombo);

    //clear display:
    sendData(clearLcdCommand,systemCommandPinCombo);

    //lastly, set the entry mode
    sendData(setEntryModeCommand + cursorIncreases + displayNotShifted, systemCommandPinCombo);

    //Turn LCD back on
    sendData(setDisplayCommand + turnDisplayOn + turnCursorOff + noBlinkCursor, systemCommandPinCombo);

    //LCD can now be written to.
    return;
}

/*
 * Shift in a byte of data into the shift register.
 * @param data
 */
void shiftIn(byte data)
{
    byte index;
    
    //Note: a 0->1 transition causes
    //shift register to shift.
    for(index = 0; index < 8; index++)
    {
        //get value on PORTC, clear out the bottom two
        //bits, then set them as desired.
        //clk = 0
        PORTC = (PINC & 0xFC) + (data & 0x01);

        //clk = 1
        PORTC = PINC + 0x02;

        data >>= 1;
    }
}




/*
 * Clears the LCD display by clearing out
 * every character on the display.
 */
void clearLcd()
{
    sendData(clearLcdCommand,systemCommandPinCombo);
    return;
}

/*
 * For sending the characters over to the LCD, if a null
 * character is being sent, it must be converted to
 * a space for the display.
 * @param line - line to convert nulls to spaces in
 */
 void convNullsToSpaces(char line[charactersPerLine+1])
{
    int index;
    for(index = 0; index < charactersPerLine; index++)
    {
        if(line[index] == '\0')
        {
            line[index] = ' ';
        }
    }
    return;
}

/*
 * (To save in code space), this method will
 * shift in data to the shift register, then
 * using the shift register send that data to 
 * the LCD.
 * @param data - data to send to LCD
 * @param linesToRaise - what bit pattern should
 * be on PORTC to do the communication
 * (ie, the function that is being executed).
 */
void sendData(byte data, byte linesToRaise)
{
    shiftIn(data);
    PORTC = (PINC & clearLcdBits) + linesToRaise;
    PORTC = PINC + enable;
    PORTC = PINC & clearLcdBits;

    //wait for the command to finish
    sleep(1);
}

/*
 * Writes 20 characters to the LCD starting at
 * the LCD's current address.
 * @param line - line to write
 */
void writeToLcd(char line[charactersPerLine+1])
{
    convNullsToSpaces(line);
    int index;
    for(index = 0; index < charactersPerLine; index++)
    {
        sendData(line[index],writeCommandPinCombo);
    }
}
/*
 * This sets line 1 of the LCD display to a certain
 * character string. Each line holds 20 characters.
 * Each characters must be specified. The last character
 * is assumed to be a null, but in reality it is ignored.
 */
void setLcdLine1(char line [charactersPerLine + 1])
{
    sendData(line1Addr,setAddrCommandPinCombo);
    writeToLcd(line);
}


/*
 * This sets line 2 of the LCD display to a certain
 * character string. Each line holds 20 characters.
 * Each characters must be specified. The last character
 * is assumed to be a null, but in reality it is ignored.
 */
void setLcdLine2(char line [charactersPerLine + 1])
{
    sendData(line2Addr,setAddrCommandPinCombo);
    writeToLcd(line);
}

/*
 * This sets line 1 of the LCD display to a certain
 * character string. Each line holds 20 characters.
 * Each characters must be specified. The last character
 * is assumed to be a null, but in reality it is ignored.
 */
void setLcdLine3(char line [charactersPerLine + 1])
{
    sendData(line3Addr,setAddrCommandPinCombo);
    writeToLcd(line);
} 

/*
 * This sets line 1 of the LCD display to a certain
 * character string. Each line holds 20 characters.
 * Each characters must be specified. The last character
 * is assumed to be a null, but in reality it is ignored.
 */
void setLcdLine4(char line [charactersPerLine + 1])
{
    sendData(line4Addr,setAddrCommandPinCombo);
    writeToLcd(line);
} 

/*
 * Creates a debug line for the display that will convert
 * the variable to a string, then combine the varName field
 * and the variable. The result will look something like this:
 * varName=value
 * If the combonation of the buffer and varName is more than
 * 20 characters, the end of the variable's value will be
 * concatinated, and the last character in the row will be '_'.
 * @param buffer - Where the result will be stored
 * @param varName - the name of the variable. The end of the name
 *                  should be marked with a null.
 * @param variable - the variable that will be converted to a
 *                   string. You can type cast bytes for this 
 *                   field.
 * @param numberBase - the number base the string will be in.
 *                     common bases are 2, 10, and 16. In the
 *                     string, binary will be denoted by 0b, 
 *                     and hex will be denoted by 0x
 */
void createDebugLine(char buffer [charactersPerLine + 1], char varName [maxVarNameSize + 1], int variable, int numberBase)
{

    //the maximum size, if using binary, is 16 bits for an int + null
    char varAsString [17];
    varAsString[0] = '\0';

    //total length is varName plus varAsString + '='
    int lengthOfVarAsString = strlen(varAsString);
    int lengthOfVarName = strlen(varName);

    //if the length of the variable name is too long,
    //it is truncated
    if(lengthOfVarName > maxVarNameSize)
    {
        varName[maxVarNameSize] = '\0';
        lengthOfVarName = 15;
    }
    int totalLength = strlen(varName) + lengthOfVarAsString + 1;

    if(numberBase == 2 || numberBase == 16)
    {

        //total length includes 0b or =0x
        totalLength += 2;
    }

    if(totalLength > charactersPerLine)
    {
        byte lengthOver = totalLength - charactersPerLine;

        varAsString[lengthOfVarAsString - lengthOver - 1] = '_';
        varAsString[lengthOfVarAsString - lengthOver]     = '\0';
    }

    //reset all characters in the buffer first
    int index;
    for(index = 0; index <= charactersPerLine; index++)
    {
        buffer[index] = '\0';
    }

    strcat(buffer,varName);

    if(numberBase == 2)
    {
        strcat(buffer,"=0b");
    }
    else if(numberBase == 16)
    {
        strcat(buffer,"=0x");
    }
    else
    {
        strcat(buffer,"=");
    }

    strcat(buffer,varAsString);

    return;
}


/*
 * All the string tables are compressed based on the largest
 * string in that table. This method will convert the string
 * to the correct string size.
 * @param input - the input string to format to the correct 
 *        size
 * @param inputSize - the size of the input string
 * @param output - an already allocated output buffer of 
 *        21 characters that the resultant string will be
 *        stored in.
 */
void formTo21Chars(char * input, int inputSize, char * output)
{
    int index;

    if(inputSize > 20)
    {
        inputSize = 20;
    }

    for(index = 0; index < inputSize; index++)
    {
        output[index] = input[index];
    }

    for(;index < maxCharsPlusNull; index++)
    {
        output[index] = '\0';
    }
}


/*
 * Formats the input string to the correct length, then
 * outputs it to the desired LCD line. Invalid lines
 * are ignored
 * @param toWrite - the line to write
 * @param size - the size of the line including null
 * @param line - line to write to
 */
void formatAndSendToLcd(int line, char * toWrite, int size)
{
    char outputBuffer [21];

    formTo21Chars(toWrite,size,outputBuffer);

    switch(line)
    {
        case 1:
            setLcdLine1(outputBuffer);
            break;
        case 2:
            setLcdLine2(outputBuffer);
            break;
        case 3:
            setLcdLine3(outputBuffer);
            break;
        case 4:
            setLcdLine4(outputBuffer);
            break;
        default:
            break;
    }
}


/*
 * Find the size of the string toWrite then sends it
 * over for formatting and sending.
 * @param line - what line to write to
 * @param toWrite - line to write
 */
void sendToLcd(int line, char * toWrite)
{
    formatAndSendToLcd(line,toWrite,strlen(toWrite));
}
